# ---- scripts for DRG -------
"""
Implements the DRG framework
"""
class DRG(nn.Module):
    """
    Main class for DRG.
    It consists of three main parts:
    - backbone
    - spatial branch of human centric and object centric: takes the features + computes
    detections from it.
    - human and object heads: takes the features + computes
        detections from it.
    """

    def __init__(self, cfg):
        super(DRG, self).__init__()

        self.backbone = build_backbone(cfg)
        self.sp_branch_human = build_spatial(cfg, self.backbone.out_channels)
        self.sp_branch_object = build_spatial_object_centric(cfg, self.backbone.out_channels)
        self.human_head = build_human_head(cfg, self.backbone.out_channels)
        self.object_head = build_object_head(cfg, self.backbone.out_channels)

    def forward(self, images, blobs, split='train'):
        """
        Arguments:
            images (list[Tensor] or ImageList): images to be processed
            blobs (dict[Tensor]): containing boxes, word embeddings, labels, masks

        Returns:
            result (list[BoxList] or dict[Tensor]): the output from the model.
                During training, it returns a dict[Tensor] which contains the losses
                and a dict[Tensor] containing score summaries for Tensorboard.
                During testing, it returns tuple[Tensor] scores of each branch and merged scores.

        """
        if self.training or split == 'val':
            label_h = blobs['human_labels']
            label_o = blobs['object_labels']
            label_ho = blobs['ho_pair_labels']
            label_ho_object = blobs['ho_pair_labels_object_centric']
            ho_mask = blobs['mask_ho']
            ho_mask_object = blobs['mask_ho_object_centric']
        else:
            label_h = None
            label_o = None
            label_ho = None
            label_ho_object = None
            ho_mask = None
            ho_mask_object = None

        features = self.backbone(images.tensors)

        h_boxes = blobs['human_boxes']
        o_boxes = blobs['object_boxes']
        pos_num = blobs['pos_num']
        object_word_embedding = blobs['object_word_embeddings']
        spatial = blobs['spatials']
        object_word_embedding_object = blobs['object_word_embeddings_object_centric']
        spatial_object = blobs['spatials_object_centric']

        human_class_logits, human_class_prob, cross_entropy_H = self.human_head(features, h_boxes, label_h, pos_num, split=split)
        object_class_logits, object_class_prob, cross_entropy_O = self.object_head(features, o_boxes, label_o, pos_num, ho_mask, split=split)
        spatial_human_class_logits, spatial_human_class_prob, cross_entropy_sp_human = self.sp_branch_human(spatial, object_word_embedding, label_ho, ho_mask, split=split)
        spatial_object_class_logits, spatial_object_class_prob, cross_entropy_sp_object = self.sp_branch_object(spatial_object, object_word_embedding_object, label_ho_object, ho_mask_object, split=split)

        if self.training or split == 'val':
            losses = {}
            losses['h_loss'] = cross_entropy_H
            losses['o_loss'] = cross_entropy_O
            losses['sp_human_loss'] = cross_entropy_sp_human
            losses['sp_object_loss'] = cross_entropy_sp_object
            total_loss = 0.1 * cross_entropy_H + 0.1 * cross_entropy_O + cross_entropy_sp_human + cross_entropy_sp_object
            losses['total_loss'] = total_loss

            score_summaries = {}
            score_summaries['human_class_prob'] = human_class_prob
            score_summaries['object_class_prob'] = object_class_prob
            score_summaries['spatial_human_class_prob'] = spatial_human_class_prob
            score_summaries['spatial_object_class_prob'] = spatial_object_class_prob

            return losses, score_summaries

        return human_class_prob, object_class_prob, spatial_human_class_prob, spatial_object_class_prob

# ----- scripts for human head -----
@registry.HUMAN_PREDICTOR.register("FPNPredictor")
class FPNPredictor(nn.Module):
    def __init__(self, cfg, in_channels):
        super(FPNPredictor, self).__init__()
        num_classes = cfg.DATASETS.NUM_CLASSES
        representation_size = in_channels

        self.cls_score = nn.Linear(representation_size, num_classes)

        nn.init.normal_(self.cls_score.weight, std=0.01)
        nn.init.constant_(self.cls_score.bias, 0)

    def forward(self, x):
        if x.ndimension() == 4:
            assert list(x.shape[2:]) == [1, 1]
            x = x.view(x.size(0), -1)
        scores = self.cls_score(x)
        return scores

def make_human_predictor(cfg, in_channels):
    func = registry.HUMAN_PREDICTOR[cfg.MODEL.ROI_BOX_HEAD.PREDICTOR]
    return func(cfg, in_channels)


class HumanHead(torch.nn.Module):
    def __init__(self, cfg, in_channels):
        super(HumanHead, self).__init__()
        self.feature_extractor = make_human_feature_extractor(cfg, in_channels)
        self.predictor = make_human_predictor(
            cfg, self.feature_extractor.out_channels)

        self.num_classes = cfg.DATASETS.NUM_CLASSES


    def forward(self, features, boxes, label_H=None, pos_num=1, split='train'):
        """
        Arguments:
            features (list[Tensor]): feature-maps from possibly several levels
            boxes (list[BoxList]): boxes
            label_H (list[BoxList], optional): the ground-truth targets.

        Returns:
            During training, returns the losses and scores for the head.
            During testing, returns detected scores.
        """

        x = self.feature_extractor(features, boxes)
        class_logits = self.predictor(x)
        class_prob = torch.sigmoid(class_logits)

        if self.training or split == 'val':
            criterion = nn.BCEWithLogitsLoss(reduction='none')

            cross_entropy_H = criterion(class_logits[:pos_num, :].view(-1, self.num_classes),
                                        label_H[:pos_num, :])
            cross_entropy_H = cross_entropy_H.mean()
            return class_logits, class_prob, cross_entropy_H
        else:
            return class_logits, class_prob, {}

# --- scripts for object head ----
@registry.OBJECT_PREDICTOR.register("FPNPredictor")
class FPNPredictor(nn.Module):
    def __init__(self, cfg, in_channels):
        super(FPNPredictor, self).__init__()
        num_classes = cfg.DATASETS.NUM_CLASSES
        representation_size = in_channels

        self.cls_score = nn.Linear(representation_size, num_classes)

        nn.init.normal_(self.cls_score.weight, std=0.01)
        nn.init.constant_(self.cls_score.bias, 0)

    def forward(self, x):
        if x.ndimension() == 4:
            assert list(x.shape[2:]) == [1, 1]
            x = x.view(x.size(0), -1)
        scores = self.cls_score(x)
        return scores

def make_object_predictor(cfg, in_channels):
    func = registry.OBJECT_PREDICTOR[cfg.MODEL.ROI_BOX_HEAD.PREDICTOR]
    return func(cfg, in_channels)


class ObjectHead(torch.nn.Module):
    def __init__(self, cfg, in_channels):
        super(ObjectHead, self).__init__()
        self.feature_extractor = make_object_feature_extractor(cfg, in_channels)
        self.predictor = make_object_predictor(
            cfg, self.feature_extractor.out_channels)

        self.num_classes = cfg.DATASETS.NUM_CLASSES


    def forward(self, features, boxes, label_HO=None, pos_num=1, HO_mask=None, split='train'):
        """
        Arguments:
            features (list[Tensor]): feature-maps from possibly several levels
            boxes (list[BoxList]): boxes
            label_HO (list[BoxList], optional): the ground-truth targets.

        Returns:
            During training, returns the losses and scores for the head.
            During testing, returns detected scores.
        """

        x = self.feature_extractor(features, boxes)
        class_logits = self.predictor(x)
        class_prob = torch.sigmoid(class_logits)

        if self.training or split == 'val':
            criterion = nn.BCEWithLogitsLoss(reduction='none')

            cross_entropy_O = criterion(class_logits[:pos_num, :].view(-1, self.num_classes),
                                        label_HO[:pos_num, :]) * HO_mask[:pos_num, :]
            cross_entropy_O = cross_entropy_O.mean()
            return class_logits, class_prob, cross_entropy_O
        else:
            return class_logits, class_prob, {}

# ----- scripts for build spatial human centric ------
class SPModule(torch.nn.Module):
    """
    Module for spatial branch computation.
    """
    def __init__(self, cfg, in_channels):
        super(SPModule, self).__init__()

        self.cfg = cfg.clone()
        self.num_classes = cfg.DATASETS.NUM_CLASSES
        if cfg.MODEL.USE_POSE:
            self.sp_to_head = nn.Sequential(
                nn.Conv2d(3, 64, 5),
                nn.MaxPool2d(2),
                nn.Conv2d(64, 32, 5),
                nn.MaxPool2d(2),
                Flatten()
            )
        else:
            self.sp_to_head = nn.Sequential(
                nn.Conv2d(2, 64, 5),
                nn.MaxPool2d(2),
                nn.Conv2d(64, 32, 5),
                nn.MaxPool2d(2),
                Flatten()
            )

        sp_head_channels = 5408 + 300
        self.sp_head_channels = sp_head_channels
        self.ssar_key_net = nn.Linear(sp_head_channels, 1024, bias=None)
        self.ssar_value_net = nn.Linear(sp_head_channels, sp_head_channels, bias=None)
        self.ssar_attn_dropout = nn.Dropout(p=0.7, inplace=True)
        self.ssar_layernorm = nn.LayerNorm(sp_head_channels)

        self.sp_head_to_tail = nn.Sequential(
            nn.Linear(sp_head_channels, 1024),
            nn.Dropout(p=0.5, inplace=True),
            nn.Linear(1024, 1024),
            nn.Dropout(p=0.5, inplace=True)
        )
        self.cls_score_net_sp = nn.Linear(1024, self.num_classes)

        self.init_weights()
        self.criterion = nn.BCEWithLogitsLoss(reduction='none')


    def init_weights(self):
        def normal_init(m, mean, stddev):
            m.weight.data.normal_(mean, stddev)
            if m.bias is not None:
                m.bias.data.zero_()

        for layer in self.sp_to_head:
            if hasattr(layer, 'weight'):
                normal_init(layer, 0, 0.01)
        normal_init(self.cls_score_net_sp, 0, 0.01)
        normal_init(self.ssar_key_net, 0, 0.01)
        self.ssar_value_net.weight.data.copy_(torch.eye(self.sp_head_channels))

        for layer in self.sp_head_to_tail:
            if hasattr(layer, 'weight'):
                normal_init(layer, 0, 0.01)

    def refinement(self, net_human_embed):
        key_embed = self.ssar_key_net(net_human_embed)
        query_embed = self.ssar_key_net(net_human_embed)
        value_embed = self.ssar_value_net(net_human_embed)

        weights = torch.matmul(query_embed, torch.t(key_embed))
        weights = weights / math.sqrt(key_embed.size(-1))
        softmax_attention = nn.Softmax(dim=1)(weights)

        weights = torch.matmul(softmax_attention, value_embed)
        weights = F.relu(weights)
        weights = self.ssar_attn_dropout(weights)
        weights = self.ssar_layernorm(net_human_embed + weights)

        return weights

    def forward(self, spatial, object_word_embedding, label_ho=None, ho_mask=None, split='train'):
        """
        Arguments:
            spatial (Tensor): spatial features of human-object-pairs for which we want to compute the predictions
            object_word_embedding (list[Tensor]): object_word_embedding of human-object-pairs

        Returns:
            During training, returns the losses and scores for the head.
            During testing, returns detected scores.
        """
        sp = self.sp_to_head(spatial) # [4, 5408]
        net_word_sp = torch.cat((object_word_embedding, sp), 1)

        X_1 = self.refinement(net_word_sp)
        X_2 = self.refinement(X_1)

        fc9_sp = self.sp_head_to_tail(X_2)

        cls_score_sp = self.cls_score_net_sp(fc9_sp)
        cls_prob_sp = torch.sigmoid(cls_score_sp)

        if self.training or split == 'val':
            cross_entropy_sp = self.criterion(cls_score_sp.view(-1, self.num_classes),
                                              label_ho) * ho_mask
            cross_entropy_sp = cross_entropy_sp.mean()
            return cls_score_sp, cls_prob_sp, cross_entropy_sp
        else:
            return cls_score_sp, cls_prob_sp, {}


# ----- scripts for build spatial object centric -------

class SPModuleObjectCentric(torch.nn.Module):
    """
    Module for spatial branch computation.
    """

    def __init__(self, cfg, in_channels):
        super(SPModuleObjectCentric, self).__init__()

        self.cfg = cfg.clone()
        self.num_classes = cfg.DATASETS.NUM_CLASSES
        self.block_number = cfg.MODEL.SSAR_BLOCK_NUMBER
        if cfg.MODEL.USE_POSE:
            self.sp_to_head = nn.Sequential(
                nn.Conv2d(3, 64, 5),
                nn.MaxPool2d(2),
                nn.Conv2d(64, 32, 5),
                nn.MaxPool2d(2),
                Flatten()
            )
        else:
            self.sp_to_head = nn.Sequential(
                nn.Conv2d(2, 64, 5),
                nn.MaxPool2d(2),
                nn.Conv2d(64, 32, 5),
                nn.MaxPool2d(2),
                Flatten()
            )

        sp_head_channels = 5408 + 300
        self.sp_head_channels = sp_head_channels
        self.ssar_key_net = nn.Linear(sp_head_channels, 1024, bias=None)
        self.ssar_value_net = nn.Linear(sp_head_channels, sp_head_channels, bias=None)
        self.ssar_attn_dropout = nn.Dropout(p=0.7, inplace=True)
        self.ssar_layernorm = nn.LayerNorm(sp_head_channels)

        self.sp_head_to_tail = nn.Sequential(
            nn.Linear(sp_head_channels, 1024),
            nn.Dropout(p=0.5, inplace=True),
            nn.Linear(1024, 1024),
            nn.Dropout(p=0.5, inplace=True)
        )
        self.cls_score_net_sp = nn.Linear(1024, self.num_classes)

        self.init_weights()
        self.criterion = nn.BCEWithLogitsLoss(reduction='none')


    def init_weights(self):
        def normal_init(m, mean, stddev):
            m.weight.data.normal_(mean, stddev)
            if m.bias is not None:
                m.bias.data.zero_()

        for layer in self.sp_to_head:
            if hasattr(layer, 'weight'):
                normal_init(layer, 0, 0.01)
        normal_init(self.cls_score_net_sp, 0, 0.01)
        normal_init(self.ssar_key_net, 0, 0.01)
        self.ssar_value_net.weight.data.copy_(torch.eye(self.sp_head_channels))

        for layer in self.sp_head_to_tail:
            if hasattr(layer, 'weight'):
                normal_init(layer, 0, 0.01)

    def refinement(self, net_human_embed):
        key_embed = self.ssar_key_net(net_human_embed)
        query_embed = self.ssar_key_net(net_human_embed)
        value_embed = self.ssar_value_net(net_human_embed)

        weights = torch.matmul(query_embed, torch.t(key_embed))
        weights = weights / math.sqrt(key_embed.size(-1))
        softmax_attention = nn.Softmax(dim=1)(weights)

        weights = torch.matmul(softmax_attention, value_embed)
        weights = F.relu(weights)
        weights = self.ssar_attn_dropout(weights)
        weights = self.ssar_layernorm(net_human_embed + weights)

        return weights


    def forward(self, spatial, object_word_embedding, label_ho=None, ho_mask=None, split='train'):
        """
        Arguments:
            spatial (Tensor): spatial features of human-object-pairs for which we want to compute the predictions
            object_word_embedding (list[Tensor]): object_word_embedding of human-object-pairs

        Returns:
            During training, returns the losses and scores for the head.
            During testing, returns detected scores.
        """
        sp = self.sp_to_head(spatial)
        net_word_sp = torch.cat((object_word_embedding, sp), 1)

        X_1 = self.refinement(net_word_sp)
        X_2 = self.refinement(X_1)

        fc9_sp = self.sp_head_to_tail(X_2)

        cls_score_sp = self.cls_score_net_sp(fc9_sp)
        cls_prob_sp = torch.sigmoid(cls_score_sp)

        if self.training or split == 'val':
            cross_entropy_sp = self.criterion(cls_score_sp.view(-1, self.num_classes),
                                              label_ho) * ho_mask
            cross_entropy_sp = cross_entropy_sp.mean()
            return cls_score_sp, cls_prob_sp, cross_entropy_sp
        else:
            return cls_score_sp, cls_prob_sp, {}
